﻿//
// PICプログラマ
//
//
// ChaNさん（http://elm-chan.org/index_j.html）の PIC24SPのソースファイルを参考にさせていただきました。
//
// MICROCHIP社 http://www.microchip.co.jp/ の
// PIC16F87_88 FLASH Memory Programming Specification 39607c
// PIC24FJXXXGA0XX Flash Programming Specification 39768d
// を頑張って読みました。が、英語が苦手なのであんまりよくわかっていません。間違いがあるかも。
//
// 2011年3月24日作成
// 2011年3月29日 PIC24FJ64GA004書き込み動作良好
// 2011年3月30日 細かい処理を別のクラスに追い出した
// 2011年3月31日 PIC16F88の書き込み動作良好。デバイスリストファイルを修正すれば他のPIC16も書けるかも
// http://kyoutan.jpn.org

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO.Ports;  // SerialPort.GetPortNames()を使うために追加
using System.IO;    // 追加 ファイル操作で使用

namespace WindowsFormsApplication1
{
    public partial class Form1 : Form
    {
        picctl pic = new picctl();

        string[] devicelist;
        uint deviceid, prgrom = 0, eraseblk = 0, writeblk = 0, eeprom = 0;
        string devicename = "未定義のデバイス";

        public Form1()
        {
            InitializeComponent();
            pic.form1 = this;         // picctlからputmsgが使えるように追加。子 Form2 の参照ポインタを設定

            try
            {
                comboBox1.Sorted = true;
                comboBox1.Items.AddRange(SerialPort.GetPortNames());    // 有効なシリアルポート一覧を取得してコンボボックスにセット
                comboBox1.SelectedIndex = 0;
            }
            catch
            {
                putmsg("シリアルポートが見つかりませんでした");
                button1.Enabled = false;    //例外が発生した（シリアルポートがない）時、ボタンを押せなくする
            }

            try
            {
                devicelist = File.ReadAllLines("devicelist.txt");
            }
            catch
            {
                devicelist = new string[0];
                putmsg("デバイスリストファイルが開けなかった");
            }
            /*
            for (int a = 0; a < devicelist.Length; a++)
            {
                putmsg(devicelist[a]);
            }
            */
            readini();  // フォームのサイズとかを読み込む
        }

        private void button2_Click(object sender, EventArgs e)
        {
            openFileDialog1.ShowDialog();
            filename.Text = openFileDialog1.FileName;
        }

        private void button1_Click(object sender, EventArgs e)
        {
            ctlenabled(false);

            clrmsg();

            if (pic.IsOpen == false)    // オープン済みじゃないか調べる
            {
                pic.PortName = (string)comboBox1.SelectedItem;
                pic.BaudRate = 115200;
                pic.DataBits = 8;              // 8ビット
                pic.Parity = Parity.None;      // パリティ無し
                pic.StopBits = StopBits.One;
                pic.RtsEnable = Enabled;       // RTS/DTR 端子の状態をセット？
                pic.DtrEnable = Enabled;
                pic.Handshake = Handshake.None;
                pic.ReadTimeout = 500;  // 500ms
                pic.WriteTimeout = 500;
                pic.WriteBufferSize = 64;
                pic.ReadBufferSize = 2048;

                try
                {
                    pic.Open();
                }
                catch
                {
                    putmsg(pic.PortName + " を開けませんでした");
                }

                if (pic.IsOpen == true)
                {
                    putmsg(pic.PortName + " オープン");

                    if (pic.test())
                    {
                        

                        putmsg("ライタ接続良好");
                        putmsg("デバイスIDリード");
                        if (devicedetect())
                        {
                            putmsg(devicename);
                            /*
                            putmsg(prgrom.ToString("X"));
                            putmsg(eraseblk.ToString("X"));
                            putmsg(writeblk.ToString("X"));
                            putmsg(eeprom.ToString("X"));
                            */
                            if (devicename.Substring(0, 5) == "PIC16")
                            {
                                pic16write(filename.Text);
                            }
                            if (devicename.Substring(0, 5) == "PIC24")
                            {
                                pic24write(filename.Text);
                            }
                        }
                        
                        /*
                        pic.pic24enticsp();

                        // PIC24接続テスト
                        if (pic.pic24checkconnection()) putmsg("PIC24を検出した");
                        else putmsg("PIC24を探したけど見つからなかった");


                        //

                        pic.pic24exiticsp();
                        */ 
                    }
                    else
                    {
                        putmsg("ライタが見つかりませんでした");
                    }
                }

                pic.Close();
                putmsg(pic.PortName + " クローズ");
            }
            else
            {
                putmsg("シリアルポートを開いていないのに、開いたことになっている？\r\n再起動してみて");
            }

            ctlenabled(true);
        }

        private bool devicedetect()
        {
            deviceid = pic16readdevid() & 0x3FF0;

            // デバイスリストから探す
            if (devicesearch("PIC16")) return true;
            if ((deviceid > 0x0000) && (deviceid < 0x3FF0))
            {
                // PIC16のデバイスIDを読みだしたけど、リストに無かったとき
                devicename = "PIC16シリーズ（未定義）";
                prgrom = 0x1000;
                eraseblk = 0x20;
                writeblk = 0x01;
                eeprom = 0x20;
                return true;
            }

            deviceid = pic24readdevid();
            if (devicesearch("PIC24")) return true;

            return false;
        }

        private bool devicesearch(string devfamily)
        {
            for (int line = 0; line < devicelist.Length; line++)
            {
                if (devicelist[line][0] == 'P')
                {
                    if (devicelist[line].Substring(0, 5) == devfamily)
                    {
                        string[] str = devicelist[line].Split(',');
                        if (deviceid == hextoshort(str[1]))
                        {
                            devicename = str[0];
                            prgrom = (uint)hextoshort(str[2]);
                            eraseblk = (uint)hextoshort(str[3]);
                            writeblk = (uint)hextoshort(str[4]);
                            eeprom = (uint)hextoshort(str[5]);
                            return true;    // 見つかった
                        }
                    }
                }
            }
            return false; // 見つからなかった
        }

        private int hextoshort(string str)
        {
            int ret;
            if (int.TryParse(str
                , System.Globalization.NumberStyles.HexNumber
                , System.Globalization.CultureInfo.InvariantCulture
                , out ret))
            {
                return ret;
            }
            else
            {
                return -1;
            }
        }

        private bool pic16write(string filename)
        {
            bool ret = true;
            uint[] buff = new uint[0x10000];   // 多めに64kワード
            uint endpos = 0;

            if (pic.readhexfile16(filename, buff))
            {
                // データの終端を調べる
                // コンフィギュレーションワードは除外
                for (uint a = 0; a < 0x2000; a++)
                {
                    if (buff[a] < 0x3FFF) endpos = a;
                }

                buff[0x2006] = pic16readdevid();

                //dump16(0, 0, endpos + 1, buff);

                putmsg("\r\n書き込み");
                
                // プログラムメモリ書き込み
                pic.pic16enticsp(); // PC=0

                pic16pgmwrite(0, endpos, buff);

                // コンフィギュレーションメモリ(0x2000以降)書き込み

                pic.pic16enticsp();
                pic.pic16loadconfig();  // PC=0x2000
                pic16cfgwrite(0x2000, 0x2008, buff);

                putmsg("\r\nベリファイ");

                pic.pic16enticsp();
                pic.pic16loadconfig();  // PC=0x2000
                if (pic16verify(0x2000, 0x2008, buff) == 0) putmsg("\r\nコンフィギュレーションワードベリファイ OK");
                else putmsg("コンフィギュレーションワードベリファイエラー");
                
                {
                    int err;

                    pic.pic16enticsp(); // PC=0
                    err = pic16verify(0, endpos, buff);
                    
                    if (err==0) putmsg("プログラムメモリベリファイ OK");
                    else putmsg(string.Format("{0} ベリファイエラー", err));
                }
                //dump16(0x0000, 0x0000, prgrom, buff);
                //dump16(0x2000, 0x2000, 16, buff);
            }

            return ret;
        }

        // ベリファイではちゃんと消去されているかを確認するため、途中スキップしない。
        // 時間かかるけどしかたない。
        private int pic16verify(uint startpos,uint endpos,uint[] buff)
        {
            int err=0;

            for (uint a = startpos; a < (endpos + 1); a++)
            {
                uint read = pic.pic16readpgm();

                if (read != buff[a])
                {
                    putmsg(string.Format("\r\nerr {0:X4} {1:X4} -> {2:X4}", a, buff[a], read));
                    err++;
                }
                pic.pic16incaddress();
                if ((a % 4) == 0) addmsg(".");
            }
            return err;
        }

        private void pic16cfgwrite(uint startpos, uint endpos, uint[] buff)
        {
            uint a;

            for (a = startpos; a < (endpos + 1); a++)
            {
                if ((a % 128) == 0) addmsg(a.ToString("X4"));

                // 消去ブロックの先頭だったら消去コマンド
                // プログラムメモリとコンフィギュレーションメモリはブロックサイズが別？
                // 0x2000で消去コマンドを実行すると8ワードしか消去されないようだ（PIC16F88で確認）
                if ((a % 8) == 0)
                {
                    pic.pic16loadpgm(0xFFFF);
                    pic.pic16beginerase();
                    addmsg("+");
                }

                if (buff[a] != 0x3FFF) // 0x3FFFならスキップ
                {
                    pic.pic16loadpgm(buff[a]);
                    pic.pic16beginprogrammingonly();    // 1ワードごとプログラミングコマンド実行
                    addmsg(".");
                }
                pic.pic16incaddress();
            }
        }

        // 低電圧でも動くように、チップイレース・バルクイレースコマンドは使わない。
        private void pic16pgmwrite(uint startpos, uint endpos, uint[] buff)
        {
            uint a,count=0;

            for (a = startpos; a < (endpos + 1); a++)
            {
                if ((a % 128) == 0) addmsg(a.ToString("X4"));
                if ((a % eraseblk) == 0)  // 消去ブロックの先頭だったら消去コマンド
                {
                    pic.pic16loadpgm(0xFFFF);   // なにかロードしないと消去できない（PCが確定しない？）
                    pic.pic16beginerase();
                    addmsg("+");
                }

                if (buff[a] == 0x3FFF) count++; // 0x3FFFならスキップ
                else pic.pic16loadpgm(buff[a]);

                if ((((a + 1) % writeblk) == 0))   // ライトブロックの最後だったら書き込みコマンド
                {
                    if (count != writeblk)  // ライトブロック内がすべて0x3FFFならスキップ
                    {
                        pic.pic16beginprogrammingonly();
                        addmsg(".");
                    }
                    count = 0;
                }

                pic.pic16incaddress();
            }

            if ((a % writeblk) != 0) // 終了したときにライトブロックの先頭じゃない（書き込みコマンドを出していない）
            {
                pic.pic16beginprogrammingonly();
                addmsg(".");
            }
        }

        private uint pic16readdevid()
        {
            uint ret;

            pic.pic16loadconfig();
            for (int a = 0; a < 6; a++) pic.pic16incaddress();
            ret = pic.pic16readpgm();
            pic.pic16exticsp();

            return ret;
        }

        private uint pic24readdevid()
        {
            uint[] buff = new uint[2];
            // 0xFF0000 DEVID
            // 0xFF0002 DEVREV
            pic.pic24enticsp();
            pic.pic24readmemory(0xFF0000, 2, buff);
            //putmsg(buff[0].ToString("X4"));
            //putmsg(buff[1].ToString("X4"));
            pic.pic24exiticsp();

            return buff[0];
        }

        private bool pic24write(string filename)
        {

            uint[] buff = new uint[0x40000];   // 多めに256kワード
            uint endpos = 0;
            bool ret = true;

            pic.pic24enticsp();

            if (pic.readhexfile32(filename, buff))
            {
                uint size = 0, writestartaddress = 0;

                // データの終端を調べる
                for (uint a = 0; a < buff.Length; a++)
                {
                    if (buff[a] < 0xFFFFFF) endpos = a;
                }

                putmsg("\r\nチップイレース");
                pic.pic24chiperase();

                //dump32(0, 0, endpos, buff);
                //dump32(0, 0, 0x2C0/2, buff);

                putmsg("\r\n書き込み");
                for (uint a = 0; a < (endpos + 1); a++)
                {
                    if ((buff[a] < 0xFFFFFF) && (a != endpos))
                    {
                        // データあり
                        size++;
                    }
                    else
                    {
                        if (a == endpos) size++;

                        if (size != 0)
                        {
                            pic.pic24writememory(writestartaddress, (writestartaddress / 2), size, buff);
                            //dump32(writestartaddress, (writestartaddress / 2), size, buff);
                            size = 0;

                        }
                        writestartaddress = (a + 1) * 2;
                    }
                }


                // ベリファイ
                // 処理時間短縮のため、書き込んでいないところはチェックしない
                putmsg("\r\nベリファイ");
                //putmsg("endaddr=" + " 0x" + (endpos * 2).ToString("X6"));

                uint startaddress = 0;

                for (uint a = 0; a < (endpos + 1); a++)
                {
                    if ((buff[a] < 0xFFFFFF) && (a != endpos))
                    {
                        // データあり
                        size++;
                    }
                    else
                    {
                        // 最後はデータがあってもここに来る
                        if (a == endpos) size++;

                        if (size != 0)
                        {
                            uint[] readbuff = new uint[size];
                            pic.pic24readmemory(startaddress, size, readbuff);
                            for (int b = 0; b < size; b++)
                            {
                                if (readbuff[b] != buff[(startaddress >> 1) + b])
                                {
                                    // ベリファイエラー
                                    putmsg(String.Format(
                                        "err {0,6:X6}: W {1,6:X6} -> R {2,6:X6} "
                                        , (startaddress + (b * 2))
                                        , buff[(startaddress >> 1) + b]
                                        , readbuff[b]));
                                    ret = false;
                                }
                            }
                            size = 0;
                        }
                        startaddress = (a + 1) * 2;
                    }
                }
                if (ret == true) putmsg("ベリファイOK");
            }

            pic.pic24exiticsp();

            return ret;
        }

        public void clrmsg()
        {
            textBox2.Text = "";
        }

        public void putmsg(string msg)
        {
            addmsg(msg + "\r\n");
        }

        public void addmsg(string msg)
        {
            // AppendTextを使えば勝手に最後の行まで勝手にスクロールする
            textBox2.AppendText(msg);

            Application.DoEvents();
        }

        private void dump32(uint address, uint offset, uint size, uint[] buff)
        {
            for (uint a = offset; ; )
            {
                addmsg(string.Format("\r\n{0:X6} :", address));
                addmsg(string.Format(" {0:X6}", buff[a++]));
                if (a > (size + offset - 1)) break;
                addmsg(string.Format(" {0:X6}", buff[a++]));
                if (a > (size + offset - 1)) break;
                addmsg(string.Format(" {0:X6}", buff[a++]));
                if (a > (size + offset - 1)) break;
                addmsg(string.Format(" {0:X6}", buff[a++]));
                if (a > (size + offset - 1)) break;
                address += 8;
            }
            putmsg("");
        }

        private void dump16(uint address, uint offset, uint size, uint[] buff)
        {
            for (uint a = offset; ; )
            {
                addmsg(string.Format("\r\n{0:X4} :", address));
                addmsg(string.Format(" {0:X4}", buff[a++]));
                if (a > (size + offset - 1)) break;
                addmsg(string.Format(" {0:X4}", buff[a++]));
                if (a > (size + offset - 1)) break;
                addmsg(string.Format(" {0:X4}", buff[a++]));
                if (a > (size + offset - 1)) break;
                addmsg(string.Format(" {0:X4}", buff[a++]));
                if (a > (size + offset - 1)) break;
                addmsg(string.Format(" {0:X4}", buff[a++]));
                if (a > (size + offset - 1)) break;
                addmsg(string.Format(" {0:X4}", buff[a++]));
                if (a > (size + offset - 1)) break;
                addmsg(string.Format(" {0:X4}", buff[a++]));
                if (a > (size + offset - 1)) break;
                addmsg(string.Format(" {0:X4}", buff[a++]));
                if (a > (size + offset - 1)) break;
                address += 8;
            }
            putmsg("");
        }

        // フォームが閉じるとき
        // フォームの位置とサイズを保存する
        private void saveini()
        {
            string[] str = new string[6];
            str[0] = this.Location.X.ToString();
            str[1] = this.Location.Y.ToString();
            str[2] = this.Size.Width.ToString();
            str[3] = this.Size.Height.ToString();
            str[4] = (string)comboBox1.SelectedItem;    // COMポート
            str[5] = filename.Text;

            File.WriteAllLines(this.Text + ".ini", str);
        }

        // フォームの位置とサイズを読み出す
        private void readini()
        {
            try
            {
                int a;
                string[] str = File.ReadAllLines(this.Text + ".ini");
                this.StartPosition = FormStartPosition.Manual;

                for (int b = 0; b < str.Length; b++)
                {
                    if (int.TryParse(str[b], out a))
                    {
                        switch (b)
                        {
                            case 0: this.Left = a; break;
                            case 1: this.Top = a; break;
                            case 2: this.Width = a; break;
                            case 3: this.Height = a; break;
                        }
                    }

                    switch (b)
                    {
                        case 4:
                            int temp;
                            temp = comboBox1.FindString(str[b]);
                            if (temp == -1) temp = 0;
                            comboBox1.SelectedIndex = temp;
                            break;
                        case 5:
                            filename.Text = str[b];
                            break;
                    }
                }
            }
            catch { }
        }

        // 終了するときにフォームの位置やサイズなどを保存する
        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            saveini();
        }

        private void ctlenabled(bool a)
        {
            button1.Enabled = a;
            button2.Enabled = a;
            filename.Enabled = a;
            comboBox1.Enabled = a;
        }
    }
}
